@@ -3,9 +3,11 @@ import random
 import os
 import numpy as np
 from neurograd import xp, Tensor, float32, int64
-from typing import Optional, List, Tuple, Union, Callable
+from typing import Optional, List, Tuple, Union, Callable, Dict
 import glob
 from collections import deque
+from concurrent.futures import ThreadPoolExecutor
+import cv2
 
 # Try to import DALI first - this determines our capabilities
 try:
@@ -13,23 +15,14 @@ try:
     from nvidia.dali import pipeline_def, Pipeline
     import nvidia.dali.fn as fn
     import nvidia.dali.types as types
-    from nvidia.dali.plugin.pytorch import DALIGenericIterator
+    from nvidia.dali.plugin.pytorch import DALIGenericIterator, LastBatchPolicy
+    from nvidia.dali.backend import PreallocateDeviceMemory, PreallocatePinnedMemory
     DALI_AVAILABLE = True
 except ImportError:
     DALI_AVAILABLE = False
     print("INFO: NVIDIA DALI not available. Falling back to OpenCV-based implementation.")
     print("      For maximum performance, install with: pip install --extra-index-url https://developer.download.nvidia.com/compute/redist nvidia-dali-cuda120")
 
-# Always import OpenCV as fallback
-try:
-    import cv2
-    CV2_AVAILABLE = True
-except ImportError:
-    CV2_AVAILABLE = False
-    print("WARNING: OpenCV not available. Image loading will be limited.")
-
-from concurrent.futures import ThreadPoolExecutor
-
 # Image file extensions
 IMG_EXTS = (
     '.png', '.jpg', '.jpeg', '.bmp', '.gif',
@@ -86,7 +79,7 @@ class ImageFolder(Dataset):
         img_mode: str = "RGB",            # "RGB", "L", etc.
         img_normalize: bool = True,       # /255 -> float
         img_transform: callable = None,   # DALI pipeline or callable
-        target_transform: callable = None,
+        one_hot_targets: bool = True,     # Convert targets to one-hot encoding
         img_dtype=xp.float32,
         target_dtype=xp.int64,
         chw: bool = True,                 # return CxHxW if True, else HxWxC
@@ -96,7 +89,7 @@ class ImageFolder(Dataset):
         self.img_mode = img_mode
         self.img_normalize = img_normalize
         self.img_transform = img_transform
-        self.target_transform = target_transform
+        self.one_hot_targets = one_hot_targets
         self.img_dtype = img_dtype
         self.target_dtype = target_dtype
         self.chw = chw
@@ -114,11 +107,47 @@ class ImageFolder(Dataset):
         self.target_mapping = {name: i for i, name in enumerate(self.target_names)}
         self.num_classes = len(self.target_names)
         
+        # Create one-hot mapping
+        self.one_hot_mapping: Dict[int, np.ndarray] = {}
+        for class_idx in range(self.num_classes):
+            one_hot = np.zeros(self.num_classes, dtype=np.float32)
+            one_hot[class_idx] = 1.0
+            self.one_hot_mapping[class_idx] = one_hot
+        
         # Convert targets to numeric labels
         self.numeric_targets = [self.target_mapping[t] for t in self.targets]
 
         print(f"ImageFolder initialized: {len(self)} samples, {self.num_classes} classes")
 
+    def get_class_name(self, class_idx: int) -> str:
+        """Get class name from class index"""
+        if class_idx < 0 or class_idx >= len(self.target_names):
+            raise ValueError(f"Class index {class_idx} out of range [0, {len(self.target_names)-1}]")
+        return self.target_names[class_idx]
+    
+    def get_class_index(self, class_name: str) -> int:
+        """Get class index from class name"""
+        if class_name not in self.target_mapping:
+            raise ValueError(f"Class name '{class_name}' not found in dataset")
+        return self.target_mapping[class_name]
+    
+    def get_one_hot(self, class_idx: int) -> np.ndarray:
+        """Get one-hot encoding from class index"""
+        if class_idx not in self.one_hot_mapping:
+            raise ValueError(f"Class index {class_idx} not found in one-hot mapping")
+        return self.one_hot_mapping[class_idx]
+    
+    def get_class_from_one_hot(self, one_hot: np.ndarray) -> int:
+        """Get class index from one-hot encoding"""
+        if one_hot.shape != (self.num_classes,):
+            raise ValueError(f"One-hot encoding must have shape ({self.num_classes},), got {one_hot.shape}")
+        
+        class_idx = np.argmax(one_hot)
+        if one_hot[class_idx] != 1.0:
+            raise ValueError("One-hot encoding must have exactly one element set to 1.0")
+            
+        return class_idx
+
     def _collect_paths(self):
         """Collect image paths and their class labels"""
         if not os.path.exists(self.root) or not os.path.isdir(self.root):
@@ -146,10 +175,11 @@ class ImageFolder(Dataset):
                         self.images.append(p)
                         self.targets.append(cls)
 
+
     def get_dali_pipeline(self, batch_size: int, shuffle: bool = True, 
-                         device: str = "cpu", num_threads: int = 4, 
-                         seed: int = 42):
-        """Create DALI pipeline for this dataset (only when DALI is available)"""
+                          device: str = "cpu", num_threads: int = 4, 
+                          seed: int = 42):
+        """Create optimized DALI pipeline for this dataset (only when DALI is available)"""
         if not DALI_AVAILABLE:
             return None
             
@@ -157,71 +187,118 @@ class ImageFolder(Dataset):
         if isinstance(self.img_transform, Pipeline):
             return self.img_transform
 
-        @pipeline_def(batch_size=batch_size, num_threads=num_threads, device_id=0, seed=seed)
+        # Pre-compute constants
+        is_gpu = device == "gpu"
+        device_id = 0 if is_gpu else None
+        prefetch_depth = 3 if is_gpu else 2
+        h, w = self.img_shape or (224, 224)
+        is_grayscale = self.img_mode.upper() in ("L", "GRAY", "GREY", "GRAYSCALE")
+        
+        # Pre-compute normalization parameters if needed
+        imagenet_mean = [0.485 * 255, 0.456 * 255, 0.406 * 255] if self.img_normalize else None
+        imagenet_std = [0.229 * 255, 0.224 * 255, 0.225 * 255] if self.img_normalize else None
+        
+        @pipeline_def(
+            batch_size=batch_size, 
+            num_threads=num_threads, 
+            device_id=device_id, 
+            seed=seed, 
+            prefetch_queue_depth=prefetch_depth,
+            enable_memory_stats=False  # Disable for better performance
+        )
         def image_pipeline():
-            # Create file reader
-            if shuffle:
-                images, labels = fn.readers.file(
-                    files=self.images,
-                    labels=self.numeric_targets,
-                    random_shuffle=True,
-                    initial_fill=batch_size * 4,
-                    name="Reader"
+            # Create file reader with optimized buffer sizes
+            initial_fill = min(batch_size * 4, len(self.images))  # Don't exceed dataset size
+            images, labels = fn.readers.file(
+                files=self.images,
+                labels=self.numeric_targets,
+                random_shuffle=shuffle,
+                initial_fill=initial_fill,
+                read_ahead=True,  # Enable read-ahead for better I/O
+                name="Reader"
+            )
+
+            # Optimized decoding strategy
+            if is_gpu:
+                # Decode on GPU with mixed device
+                images = fn.decoders.image(
+                    images, 
+                    device="mixed",
+                    output_type=types.RGB
                 )
             else:
-                images, labels = fn.readers.file(
-                    files=self.images,
-                    labels=self.numeric_targets,
-                    random_shuffle=False,
-                    name="Reader"
+                # Decode and resize on CPU
+                images = fn.decoders.image(
+                    images,
+                    device="cpu", 
+                    output_type=types.RGB
                 )
-
-            # Decode and apply RandomResizedCrop
-            h, w = self.img_shape if self.img_shape is not None else (224, 224)
-            images = fn.decoders.image_random_crop(
-                images,
-                device=device,
-                output_type=types.RGB,
-                random_aspect_ratio=[0.8, 1.25],
-                random_area=[0.1, 1.0],
-                num_attempts=100
-            )
-            
-            # Resize the cropped image to the final desired shape
             images = fn.resize(
                 images,
-                device=device,
-                size=[h, w],
+                resize_x=w,
+                resize_y=h,
                 interp_type=types.INTERP_LINEAR
             )
 
-            # Handle grayscale conversion if needed
-            if self.img_mode.upper() in ("L", "GRAY", "GREY", "GRAYSCALE"):
+            # Handle grayscale conversion early if needed
+            if is_grayscale:
                 images = fn.color_space_conversion(
-                    images, device=device, image_type=types.RGB, output_type=types.GRAY
+                    images, 
+                    image_type=types.RGB, 
+                    output_type=types.GRAY
                 )
 
-            # Apply remaining augmentations
+            # Apply custom transforms if provided (with error handling)
             if self.img_transform and callable(self.img_transform):
                 try:
                     images = self.img_transform(images)
                 except Exception as e:
-                    print(f"WARNING: Failed to apply DALI callable transform. Error: {e}. Skipping transform.")
+                    # Log warning but continue - could use logging instead of print
+                    print(f"WARNING: DALI transform failed: {e}. Skipping transform.")
 
-            # Normalize to [0, 1] if requested
+            # Optimized normalization and formatting
             if self.img_normalize:
-                images = fn.cast(images, dtype=types.FLOAT)
-                images = images / 255.0
+                # Single operation for normalize + layout conversion
+                images = fn.crop_mirror_normalize(
+                    images,
+                    dtype=types.FLOAT,
+                    output_layout="CHW" if self.chw else "HWC",
+                    mean=imagenet_mean,
+                    std=imagenet_std,
+                    crop=(h, w),  # Explicit crop size
+                    mirror=0  # Disable mirroring for inference
+                )
             else:
+                # Efficient cast and transpose
                 images = fn.cast(images, dtype=types.FLOAT)
-
-            # Transpose to CHW format if requested
-            if self.chw:
-                images = fn.transpose(images, device=device, perm=[2, 0, 1])
+                if self.chw:
+                    images = fn.transpose(images, perm=[2, 0, 1])
             
-            return images, labels.gpu() if device == "gpu" else labels
+            # Efficient one-hot encoding
+            if self.one_hot_targets:
+                # Use int32 for better performance than default int64
+                labels = fn.cast(labels, dtype=types.INT32)
+                labels = fn.one_hot(labels, num_classes=self.num_classes)
+            
+            # Move to GPU at the end to minimize transfers
+            if is_gpu:
+                images = images.gpu()
+                labels = labels.gpu()
+            
+            return images, labels
 
-        return image_pipeline()
+        # Build and return the pipeline
+        pipeline = image_pipeline()
+        
+        # Optional: warm up the pipeline for consistent timing
+        # This can be enabled based on a parameter
+        try:
+            pipeline.build()
+        except Exception as e:
+            print(f"WARNING: Pipeline build failed: {e}")
+            return None
+            
+        return pipeline
 
     def _apply_img_transform(self, arr: np.ndarray) -> np.ndarray:
         """Apply image transforms (OpenCV fallback path)"""
@@ -251,9 +328,6 @@ class ImageFolder(Dataset):
 
     def _load_image_opencv(self, path: str) -> np.ndarray:
         """Load image using OpenCV (fallback implementation)"""
-        if not CV2_AVAILABLE:
-            raise RuntimeError("OpenCV not available for image loading")
-            
         # OpenCV-only fast decode/resize
         mode = (self.img_mode or "RGB").upper()
         if mode in ("L", "GRAY", "GREY", "GRAYSCALE"):
@@ -310,11 +384,15 @@ class ImageFolder(Dataset):
         # Load image using OpenCV
         image = self._load_image_opencv(img_path)
         
-        # Apply target transform
-        if self.target_transform:
-            target = self.target_transform(target)
+        # Apply one-hot encoding if requested
+        if self.one_hot_targets:
+            # Use our precomputed one-hot mapping
+            target = self.one_hot_mapping[target]
+            target_dtype = float32  # One-hot targets should be float
+        else:
+            target_dtype = self.target_dtype
         
-        return Tensor(image, dtype=self.img_dtype), Tensor(target, dtype=self.target_dtype)
+        return Tensor(image, dtype=self.img_dtype), Tensor(target, dtype=target_dtype)
 
     def shuffle(self, seed: Optional[int] = None):
         """Shuffle the dataset"""
@@ -349,7 +427,8 @@ class ImageFolder(Dataset):
         return (f"ImageFolder(root='{self.root}', samples={len(self)}, "
                 f"classes={self.num_classes}, "
                 f"shape={shape}, img_dtype={self.img_dtype}, target_dtype={self.target_dtype}, "
-                f"mode='{self.img_mode}', normalize={self.img_normalize}, chw={self.chw})")
+                f"mode='{self.img_mode}', normalize={self.img_normalize}, chw={self.chw}, "
+                f"one_hot_targets={self.one_hot_targets})")
 
 
 class DataLoader:
@@ -381,11 +460,19 @@ class DataLoader:
         self._dali_iter = None
         self._executor: Optional[ThreadPoolExecutor] = None
         
+        # Preallocate memory for better performance
+        if DALI_AVAILABLE:
+            try:
+                if self.device == "gpu":
+                    PreallocateDeviceMemory(int(0.5 * 1024**3), 0)  # 0.5GB GPU memory
+                PreallocatePinnedMemory(int(0.25 * 1024**3))  # 0.25GB pinned memory
+            except Exception as e:
+                print(f"WARNING: Memory preallocation failed: {e}")
+        
         # Determine if we should use DALI
         self.use_dali = (
             DALI_AVAILABLE and 
-            isinstance(dataset, ImageFolder) and 
-            self.device in ["gpu", "cuda"]
+            isinstance(dataset, ImageFolder)
         )
         
         if self.use_dali:
@@ -401,12 +488,18 @@ class DataLoader:
         if not self.use_dali or not isinstance(self.dataset, ImageFolder):
             return
         
+        # Device-specific thread optimization
+        if self.device == "gpu":
+            num_threads = min(4, self.num_workers)  # Fewer threads for GPU
+        else:
+            num_threads = self.num_workers
+        
         # Create pipeline
         self._pipeline = self.dataset.get_dali_pipeline(
             batch_size=self.batch_size,
             shuffle=self.shuffle,
             device=self.device,
-            num_threads=self.num_workers,
+            num_threads=num_threads,
             seed=self.seed
         )
         
@@ -414,16 +507,20 @@ class DataLoader:
             # Build the pipeline
             self._pipeline.build()
             
+            # Set appropriate last batch policy (FIXED: use enum instead of string)
+            last_batch_policy = LastBatchPolicy.DROP if self.drop_last else LastBatchPolicy.PARTIAL
+            
             # Create DALI iterator
-            last_batch_policy = 'drop' if self.drop_last else 'part'
             self._dali_iter = DALIGenericIterator(
                 self._pipeline, 
                 output_map=["images", "labels"],
                 last_batch_policy=last_batch_policy,
-                auto_reset=True
+                auto_reset=False,
+                reader_name="Reader",
+                prepare_first_batch=True  # Preload first batch
             )
             print(f"DALI pipeline initialized: batch_size={self.batch_size}, device={self.device}, "
-                  f"num_threads={self.num_workers}")
+                  f"num_threads={num_threads}")
 
     def __len__(self):
         n = len(self.dataset)
